searchNodes=[{"type":"module","title":"Schematic","doc":"schematic is a library for data specification, validation, and transformation. schematic works by constructing schematics that specify your data and can then unify to them from external data and dump your internal data back to the external data. There are 12 builtin schematics that you can use to build new schematics that fit your own domain model. bool/0 str/0 atom/0 int/0 float/0 list/1 tuple/1 map/1 schema/2 keyword/1 raw/2 any/0 all/1 oneof/1 Literals can be used as schematics and are unified with == semantics. Struct literals can be used as schematics, and the input is unified by seeing if it is an instance of the given struct. Example Let's take a look at an example schematic for a JSON-RPC request for a bookstore API. defmodule Bookstore do defmodule Datetime do import Schematic def schematic ( ) do raw ( fn i , :to -&gt; is_binary ( i ) and match? ( { :ok , _ , _ } , DateTime . from_iso8601 ( i ) ) i , :from -&gt; match? ( % DateTime { } , i ) end , transform : fn i , :to -&gt; { :ok , dt , _ } = DateTime . from_iso8601 ( i ) dt i , :from -&gt; DateTime . to_iso8601 ( i ) end ) end end defmodule Author do import Schematic defstruct [ :name ] def schematic ( ) do schema ( __MODULE__ , %{ name : str ( ) } ) end end defmodule Book do import Schematic defstruct [ :title , :authors , :publication_date ] def schematic ( ) do schema ( __MODULE__ , %{ { &quot;publicationDate&quot; , :publication_date } =&gt; Bookstore.Datetime . schematic ( ) , title : str ( ) , authors : list ( Bookstore.Author . schematic ( ) ) } ) end end defmodule BooksListResult do import Schematic defstruct [ :books ] def schematic ( ) do schema ( __MODULE__ , %{ books : list ( Bookstore.Book . schematic ( ) ) } ) end end defmodule BooksListParams do import Schematic defstruct [ :query , :order ] def schematic ( ) do schema ( __MODULE__ , %{ query : nullable ( map ( %{ { &quot;field&quot; , :field } =&gt; oneof ( [ &quot;title&quot; , &quot;authors&quot; , &quot;publication_date&quot; ] ) , { &quot;value&quot; , :value } =&gt; str ( ) } ) ) , order : nullable ( oneof ( [ &quot;asc&quot; , &quot;desc&quot; ] ) ) } ) end end defmodule BooksList do import Schematic defstruct [ :id , :method , :params ] def schematic ( ) do schema ( __MODULE__ , %{ id : int ( ) , method : &quot;books/list&quot; , params : Bookstore.BooksListParams . schematic ( ) } ) end end end Reading external data into your data model. iex&gt; alias SchematicTest.Bookstore iex&gt; import Schematic iex&gt; unify ( Bookstore.BooksList . schematic ( ) , %{ ...&gt; &quot;id&quot; =&gt; 99 , ...&gt; &quot;method&quot; =&gt; &quot;books/list&quot; , ...&gt; &quot;params&quot; =&gt; %{ ...&gt; &quot;query&quot; =&gt; %{ ...&gt; &quot;field&quot; =&gt; &quot;authors&quot; , ...&gt; &quot;value&quot; =&gt; &quot;Michael Crichton&quot; ...&gt; } , ...&gt; &quot;order&quot; =&gt; &quot;desc&quot; ...&gt; } ...&gt; } ) { :ok , % Bookstore.BooksList { id : 99 , method : &quot;books/list&quot; , params : % Bookstore.BooksListParams { query : %{ field : &quot;authors&quot; , value : &quot;Michael Crichton&quot; } , order : &quot;desc&quot; } } } Dumping your internal data model. iex&gt; alias SchematicTest.Bookstore iex&gt; import Schematic iex&gt; dump ( Bookstore.BooksListResult . schematic ( ) , % Bookstore.BooksListResult { ...&gt; books : [ ...&gt; % Bookstore.Book { ...&gt; title : &quot;Jurassic Park&quot; , ...&gt; authors : [ % Bookstore.Author { name : &quot;Michael Crichton&quot; } ] , ...&gt; publication_date : ~U[1990-11-20 00:00:00.000000Z] ...&gt; } , ...&gt; % Bookstore.Book { ...&gt; title : &quot;The Lost World&quot; , ...&gt; authors : [ % Bookstore.Author { name : &quot;Michael Crichton&quot; } ] , ...&gt; publication_date : ~U[1995-09-08 00:00:00.000000Z] ...&gt; } ...&gt; ] ...&gt; } ) { :ok , %{ &quot;books&quot; =&gt; [ %{ &quot;authors&quot; =&gt; [ %{ &quot;name&quot; =&gt; &quot;Michael Crichton&quot; } ] , &quot;publicationDate&quot; =&gt; &quot;1990-11-20T00:00:00.000000Z&quot; , &quot;title&quot; =&gt; &quot;Jurassic Park&quot; } , %{ &quot;authors&quot; =&gt; [ %{ &quot;name&quot; =&gt; &quot;Michael Crichton&quot; } ] , &quot;publicationDate&quot; =&gt; &quot;1995-09-08T00:00:00.000000Z&quot; , &quot;title&quot; =&gt; &quot;The Lost World&quot; } ] } } Telemetry schematic fires the following events: [:schematic, :unify, :start] - Fired when unification starts. [:schematic, :unify, :stop] - Fired when unification stops. [:schematic, :unify, :exception] - Fired when unification raises an exception.","ref":"Schematic.html"},{"type":"function","title":"Schematic.all/1","doc":"Specifies that the data must unify with all of the given schematics. On error, returns a list of validation messages. If a schematic raises an exception, it is caught and the error &quot;is invalid&quot; is returned. iex&gt; schematic = all ( [ int ( ) , raw ( &amp; Kernel . &lt; ( &amp;1 , 10 ) , message : &quot;must be less than 10&quot; ) , raw ( &amp; ( Kernel . rem ( &amp;1 , 2 ) == 0 ) , message : &quot;must be divisible by 2&quot; ) ] ) iex&gt; { :ok , 8 } = unify ( schematic , 8 ) iex&gt; { :error , [ &quot;must be less than 10&quot; , &quot;must be divisible by 2&quot; ] } = unify ( schematic , 15 ) iex&gt; { :error , [ &quot;expected an integer&quot; , &quot;must be less than 10&quot; , &quot;is invalid&quot; ] } = unify ( schematic , &quot;15&quot; )","ref":"Schematic.html#all/1"},{"type":"function","title":"Schematic.any/0","doc":"Specifies that the data can be any thing. Usage iex&gt; schematic = any ( ) iex&gt; { :ok , &quot;hi!&quot; } = unify ( schematic , &quot;hi!&quot; ) iex&gt; { :ok , [ :one , :two , :three ] } = unify ( schematic , [ :one , :two , :three ] ) iex&gt; { :ok , true } = unify ( schematic , true )","ref":"Schematic.html#any/0"},{"type":"function","title":"Schematic.atom/0","doc":"Specifies that the data is an atom. Usage Any string. iex&gt; schematic = atom ( ) iex&gt; { :ok , :hi } = unify ( schematic , :hi ) iex&gt; { :error , &quot;expected an atom&quot; } = unify ( schematic , &quot;boom&quot; )","ref":"Schematic.html#atom/0"},{"type":"function","title":"Schematic.bool/0","doc":"Specifies that the data is a boolean or a specific boolean. Usage Any boolean. iex&gt; schematic = bool ( ) iex&gt; { :ok , true } = unify ( schematic , true ) iex&gt; { :ok , false } = unify ( schematic , false ) iex&gt; { :error , &quot;expected a boolean&quot; } = unify ( schematic , :boom ) A boolean literal. iex&gt; schematic = true iex&gt; { :ok , true } = unify ( schematic , true ) iex&gt; { :error , &quot;expected true&quot; } = unify ( schematic , :boom )","ref":"Schematic.html#bool/0"},{"type":"function","title":"Schematic.dump/2","doc":"Dump your internal data to their external data structures. See all the other functions for information on how to create schematics.","ref":"Schematic.html#dump/2"},{"type":"function","title":"Schematic.float/0","doc":"Specifies that the data is a float or specific float. Usage Any float iex&gt; schematic = float ( ) iex&gt; { :ok , 99.0 } = unify ( schematic , 99.0 ) iex&gt; { :error , &quot;expected a float&quot; } = unify ( schematic , :boom ) A float literal. iex&gt; schematic = 99.0 iex&gt; { :ok , 99.0 } = unify ( schematic , 99.0 ) iex&gt; { :error , ~s|expected 99.0| } = unify ( schematic , :ninetynine )","ref":"Schematic.html#float/0"},{"type":"function","title":"Schematic.int/0","doc":"Specifies that the data is an integer or a specific integer. Usage Any integer. iex&gt; schematic = int ( ) iex&gt; { :ok , 99 } = unify ( schematic , 99 ) iex&gt; { :error , &quot;expected an integer&quot; } = unify ( schematic , :boom ) A integer literal. iex&gt; schematic = 99 iex&gt; { :ok , 99 } = unify ( schematic , 99 ) iex&gt; { :error , ~s|expected 99| } = unify ( schematic , :ninetynine )","ref":"Schematic.html#int/0"},{"type":"function","title":"Schematic.keyword/0","doc":"","ref":"Schematic.html#keyword/0"},{"type":"function","title":"Schematic.keyword/1","doc":"Specifies that the data is a keyword list with the given keys (literal values) that unify to the provided blueprint. Unification errors for keys are returned in a keyword list with the key as the key and the value as the error. Keyword schematics serve as a way to permit certain keys and discard all others. Keys are non-nullable unless the value schematic is marked with nullable/1 . This allows the value of the key to be nil. Keys are considered required unless tagged with optional/1 . This allows the entire key to be absent from the source data. If the key is present, it must unify according to the given schematic. Keyword schematics allow multiple instances of the same key. Basic Usage The most basic map schematic can look like the following. iex&gt; schematic = keyword ( %{ ...&gt; league : oneof ( [ &quot;NBA&quot; , &quot;MLB&quot; , &quot;NFL&quot; ] ) , ...&gt; } ) iex&gt; # ignores the `&quot;team&quot;` key iex&gt; { :ok , [ league : &quot;NBA&quot; ] } == unify ( schematic , [ league : &quot;NBA&quot; , team : &quot;Chicago Bulls&quot; ] ) true iex&gt; { :error , ...&gt; [ ...&gt; league : ~s|expected either &quot;NBA&quot;, &quot;MLB&quot;, or &quot;NFL&quot;| ...&gt; ] } = unify ( schematic , [ league : &quot;NHL&quot; ] ) With a permissive keyword list If you want to only check that the data is a keyword list, but not the shape, you can use keyword/0 . iex&gt; schematic = keyword ( ) iex&gt; { :ok , [ league : &quot;NBA&quot; ] } = unify ( schematic , [ league : &quot;NBA&quot; ] ) With nullable/1 Marking a value as nullable using nullable/1 . This means the value of the key can be nil, but the key itself must be present. If you want to omit the key entirely, consider marking the key with optional/1 . iex&gt; schematic = keyword ( %{ ...&gt; title : str ( ) , ...&gt; description : nullable ( str ( ) ) ...&gt; } ) iex&gt; { :ok , [ title : &quot;Elixir 101&quot; , description : nil ] } = unify ( schematic , [ title : &quot;Elixir 101&quot; , description : nil ] ) iex&gt; { :ok , [ title : &quot;Elixir 101&quot; , description : &quot;An amazing Elixir class&quot; ] } = unify ( schematic , [ title : &quot;Elixir 101&quot; , description : &quot;An amazing Elixir class&quot; ] ) With optional/1 Marking a key as optional using optional/1 . This means that you can omit the key from the input and that the unified output will not contain the key if it wasn't in the input. If the key is provided, it must unify according to the given schematic. You can also provide a default value for an optional key with optional/2 . Likewise, using dump/2 will also omit that key, unless it has a default value. iex&gt; schematic = keyword ( %{ ...&gt; :title =&gt; str ( ) , ...&gt; optional ( :description ) =&gt; str ( ) , ...&gt; optional ( :kind , &quot;technology&quot; ) =&gt; str ( ) ...&gt; } ) iex&gt; { :ok , [ title : &quot;Elixir 101&quot; , description : &quot;An amazing programming course.&quot; , kind : &quot;technology&quot; ] } = unify ( schematic , [ title : &quot;Elixir 101&quot; , description : &quot;An amazing programming course.&quot; ] ) iex&gt; { :ok , [ title : &quot;Elixir 101&quot; , kind : &quot;computer science&quot; ] } = unify ( schematic , [ title : &quot;Elixir 101&quot; , kind : &quot;computer science&quot; ] ) iex&gt; { :ok , [ title : &quot;Elixir 101&quot; , kind : &quot;computer science&quot; ] } = dump ( schematic , [ title : &quot;Elixir 101&quot; , kind : &quot;computer science&quot; ] ) iex&gt; { :ok , [ title : &quot;Elixir 101&quot; , kind : &quot;technology&quot; ] } = dump ( schematic , [ title : &quot;Elixir 101&quot; ] ) :values Instead of passing a blueprint, which specifies keys and values, you can pass a :values option which provide schematic that all values in the input must unify to. iex&gt; schematic = keyword ( values : oneof ( [ str ( ) , int ( ) ] ) ) iex&gt; { :ok , [ type : &quot;big&quot; , quantity : 99 ] } = unify ( schematic , [ type : &quot;big&quot; , quantity : 99 ] ) iex&gt; { :error , [ quantity : &quot;expected either a string or an integer&quot; ] } = unify ( schematic , [ type : &quot;big&quot; , quantity : [ 99 ] ] ) Recursive Schematics See map/1 for more information on recursive schematics","ref":"Schematic.html#keyword/1"},{"type":"function","title":"Schematic.list/0","doc":"Specifies that the data is a list of any size and contains anything. Usage iex&gt; schematic = list ( ) iex&gt; { :ok , [ &quot;one&quot; , 2 , :three ] } = unify ( schematic , [ &quot;one&quot; , 2 , :three ] ) iex&gt; { :error , &quot;expected a list&quot; } = unify ( schematic , :hi )","ref":"Schematic.html#list/0"},{"type":"function","title":"Schematic.list/1","doc":"Specifies that the data is a list whose items unify to the given schematic. Lists whose elements do not unify return a list of :ok and :error tuples. Usage iex&gt; schematic = list ( oneof ( [ str ( ) , int ( ) ] ) ) iex&gt; { :ok , [ &quot;one&quot; , 2 , &quot;three&quot; ] } = unify ( schematic , [ &quot;one&quot; , 2 , &quot;three&quot; ] ) iex&gt; { :error , [ ok : &quot;one&quot; , ok : 2 , error : &quot;expected either a string or an integer&quot; ] } = unify ( schematic , [ &quot;one&quot; , 2 , :three ] )","ref":"Schematic.html#list/1"},{"type":"function","title":"Schematic.map/1","doc":"Specifies that the data is a map with the given keys (literal values) that unify to the provided blueprint. Unification errors for keys are returned in a map with the key as the key and the value as the error. Map schematics serve as a way to permit certain keys and discard all others. Keys are non-nullable unless the value schematic is marked with nullable/1 . This allows the value of the key to be nil as well as the key to be absent from the source data. Keys are considered required unless tagged with optional/1 . This allows the entire key to be absent from the source data. If the key is present, it must unify according to the given schematic. Basic Usage The most basic map schematic can look like the following. iex&gt; schematic = map ( %{ ...&gt; &quot;league&quot; =&gt; oneof ( [ &quot;NBA&quot; , &quot;MLB&quot; , &quot;NFL&quot; ] ) , ...&gt; } ) iex&gt; # ignores the `&quot;team&quot;` key iex&gt; { :ok , %{ &quot;league&quot; =&gt; &quot;NBA&quot; } } == unify ( schematic , %{ &quot;league&quot; =&gt; &quot;NBA&quot; , &quot;team&quot; =&gt; &quot;Chicago Bulls&quot; } ) true iex&gt; { :error , ...&gt; %{ ...&gt; &quot;league&quot; =&gt; ...&gt; ~s|expected either &quot;NBA&quot;, &quot;MLB&quot;, or &quot;NFL&quot;| ...&gt; } } = unify ( schematic , %{ &quot;league&quot; =&gt; &quot;NHL&quot; } ) With a permissive map If you want to only check that the data is a map, but not the shape, you can use map/0 . iex&gt; schematic = map ( ) iex&gt; { :ok , %{ &quot;league&quot; =&gt; &quot;NBA&quot; } } = unify ( schematic , %{ &quot;league&quot; =&gt; &quot;NBA&quot; } ) With nullable/1 Marking a key as nullable using nullable/1 . The key must be present for nullable values. iex&gt; schematic = map ( %{ ...&gt; &quot;title&quot; =&gt; str ( ) , ...&gt; &quot;description&quot; =&gt; nullable ( str ( ) ) ...&gt; } ) iex&gt; { :ok , %{ &quot;title&quot; =&gt; &quot;Elixir 101&quot; , &quot;description&quot; =&gt; nil } } = unify ( schematic , %{ &quot;title&quot; =&gt; &quot;Elixir 101&quot; , &quot;description&quot; =&gt; nil } ) iex&gt; { :ok , %{ &quot;title&quot; =&gt; &quot;Elixir 101&quot; , &quot;description&quot; =&gt; nil } } = dump ( schematic , %{ &quot;title&quot; =&gt; &quot;Elixir 101&quot; , &quot;description&quot; =&gt; nil } ) iex&gt; { :ok , %{ &quot;title&quot; =&gt; &quot;Elixir 101&quot; , &quot;description&quot; =&gt; &quot;A introductory Elixir lesson&quot; } } = unify ( schematic , %{ &quot;title&quot; =&gt; &quot;Elixir 101&quot; , &quot;description&quot; =&gt; &quot;A introductory Elixir lesson&quot; } ) iex&gt; { :ok , %{ &quot;title&quot; =&gt; &quot;Elixir 101&quot; , &quot;description&quot; =&gt; &quot;A introductory Elixir lesson&quot; } } = dump ( schematic , %{ &quot;title&quot; =&gt; &quot;Elixir 101&quot; , &quot;description&quot; =&gt; &quot;A introductory Elixir lesson&quot; } ) With optional/1 Marking a key as optional using optional/1 . This means that you can omit the key from the input and that the unified output will not contain the key if it wasn't in the input. If the key is provided, it must unify according to the given schematic. You can also provide a default value for an optional key with optional/2 . Likewise, using dump/2 will also omit that key, unless it has a default value. iex&gt; schematic = map ( %{ ...&gt; &quot;title&quot; =&gt; str ( ) , ...&gt; optional ( &quot;description&quot; ) =&gt; str ( ) , ...&gt; optional ( &quot;kind&quot; , &quot;technology&quot; ) =&gt; str ( ) ...&gt; } ) iex&gt; { :ok , %{ &quot;title&quot; =&gt; &quot;Elixir 101&quot; , &quot;description&quot; =&gt; &quot;An amazing programming course.&quot; , &quot;kind&quot; =&gt; &quot;technology&quot; } } = unify ( schematic , %{ &quot;title&quot; =&gt; &quot;Elixir 101&quot; , &quot;description&quot; =&gt; &quot;An amazing programming course.&quot; } ) iex&gt; { :ok , %{ &quot;title&quot; =&gt; &quot;Elixir 101&quot; , &quot;kind&quot; =&gt; &quot;computer science&quot; } } = unify ( schematic , %{ &quot;title&quot; =&gt; &quot;Elixir 101&quot; , &quot;kind&quot; =&gt; &quot;computer science&quot; } ) iex&gt; { :ok , %{ &quot;title&quot; =&gt; &quot;Elixir 101&quot; , &quot;kind&quot; =&gt; &quot;computer science&quot; } } = dump ( schematic , %{ &quot;title&quot; =&gt; &quot;Elixir 101&quot; , &quot;kind&quot; =&gt; &quot;computer science&quot; } ) iex&gt; { :ok , %{ &quot;title&quot; =&gt; &quot;Elixir 101&quot; , &quot;kind&quot; =&gt; &quot;technology&quot; } } = dump ( schematic , %{ &quot;title&quot; =&gt; &quot;Elixir 101&quot; } ) With :keys and :values Instead of passing a blueprint, which specifies keys and values, you can pass a :keys and :values options which provide schematic that all keys and values in the input must unify to. iex&gt; schematic = map ( keys : str ( ) , values : oneof ( [ str ( ) , int ( ) ] ) ) iex&gt; { :ok , %{ &quot;type&quot; =&gt; &quot;big&quot; , &quot;quantity&quot; =&gt; 99 } } = unify ( schematic , %{ &quot;type&quot; =&gt; &quot;big&quot; , &quot;quantity&quot; =&gt; 99 } ) iex&gt; { :error , %{ &quot;quantity&quot; =&gt; &quot;expected either a string or an integer&quot; } } = unify ( schematic , %{ &quot;type&quot; =&gt; &quot;big&quot; , &quot;quantity&quot; =&gt; [ 99 ] } ) Transforming Keys During unification, key transformation can be performed if it is specified in the schematic. You can specify a key as a 2-tuple with the first element being the input key and the second element being the output key. When calling dump/2 , the key will be turned from the output key back to the input key (and will also be revalidated). This is useful for transforming string keys to atom keys as well as camelCase keys to snake_case keys. Key transformation can also be used when declaring an optional key with optional/1 . iex&gt; schematic = map ( %{ ...&gt; { &quot;teamName&quot; , :team_name } =&gt; str ( ) ...&gt; } ) iex&gt; { :ok , %{ team_name : &quot;Chicago Bulls&quot; } } = unify ( schematic , %{ &quot;teamName&quot; =&gt; &quot;Chicago Bulls&quot; } ) iex&gt; { :ok , %{ &quot;teamName&quot; =&gt; &quot;Chicago Bulls&quot; } } = dump ( schematic , %{ team_name : &quot;Chicago Bulls&quot; } ) Recursive Schematics One can define schematics that specify keys whose values are themselves. For this to be possible, recursive schematics must terminate some way. This can be achieved by specifying those keys as optional/1 or within a oneof/1 schematic. Recursive schematics are specified as a MFA tuple, lazy_schematic/0 . iex&gt; defmodule Tree do ...&gt; import Schematic ...&gt; ...&gt; def schematic ( ) do ...&gt; map ( %{ values : list ( Tree . branch ( ) ) } ) ...&gt; end ...&gt; ...&gt; def branch ( ) do ...&gt; map ( %{ ...&gt; values : list ( oneof ( [ Tree . leaf ( ) , { __MODULE__ , :branch , [ ] } ] ) ) ...&gt; } ) ...&gt; end ...&gt; ...&gt; def leaf ( ) do ...&gt; map ( %{ ...&gt; value : str ( ) ...&gt; } ) ...&gt; end ...&gt; end iex&gt; input = %{ ...&gt; type : &quot;root&quot; , ...&gt; values : [ ...&gt; %{ ...&gt; type : &quot;branch&quot; , ...&gt; values : [ ...&gt; %{ ...&gt; type : &quot;leaf&quot; , ...&gt; value : &quot;i&#39;m a leaf&quot; ...&gt; } , ...&gt; %{ ...&gt; type : &quot;branch&quot; , ...&gt; values : [ ...&gt; %{ ...&gt; type : &quot;leaf&quot; , ...&gt; value : &quot;i&#39;m another leaf&quot; ...&gt; } ...&gt; ] ...&gt; } ...&gt; ] ...&gt; } ...&gt; ] ...&gt; } iex&gt; unify ( SchematicTest.Tree . schematic ( ) , input ) { :ok , %{ values : [ %{ values : [ %{ value : &quot;i&#39;m a leaf&quot; } , %{ values : [ %{ value : &quot;i&#39;m another leaf&quot; } ] } ] } ] } }","ref":"Schematic.html#map/1"},{"type":"function","title":"Schematic.map/2","doc":"Map schematics can be extended by using map/2 . iex&gt; player = map ( %{ name : str ( ) , team : str ( ) } ) iex&gt; baseball_player = map ( player , %{ home_runs : int ( ) } ) iex&gt; unify ( baseball_player , %{ name : &quot;Sammy Sosa&quot; , team : &quot;Cubs&quot; , home_runs : 609 , favorite_food : &quot;Hot Dog&quot; } ) { :ok , %{ name : &quot;Sammy Sosa&quot; , team : &quot;Cubs&quot; , home_runs : 609 } }","ref":"Schematic.html#map/2"},{"type":"function","title":"Schematic.nullable/1","doc":"Shortcut for specifiying that a schematic can be either null or the schematic. Usage iex&gt; schematic = nullable ( str ( ) ) iex&gt; { :ok , nil } = unify ( schematic , nil ) iex&gt; { :ok , &quot;hi!&quot; } = unify ( schematic , &quot;hi!&quot; ) iex&gt; { :error , &quot;expected either null or a string&quot; } = unify ( schematic , :boom )","ref":"Schematic.html#nullable/1"},{"type":"function","title":"Schematic.oneof/1","doc":"Specifies that the data unifies to one of the given schematics. Can be called with a list of schematics or a function. With a list When called with a list of schematics, they will be traversed during unification and the first one to unify will be returned. If none of them unify, then an error is returned. iex&gt; team = map ( %{ name : str ( ) , league : str ( ) } ) iex&gt; player = map ( %{ name : str ( ) , team : str ( ) } ) iex&gt; schematic = oneof ( [ team , player ] ) iex&gt; { :ok , %{ name : &quot;Indiana Pacers&quot; , league : &quot;NBA&quot; } } = unify ( schematic , %{ name : &quot;Indiana Pacers&quot; , league : &quot;NBA&quot; } ) iex&gt; { :ok , %{ name : &quot;George Hill&quot; , team : &quot;Indiana Pacers&quot; } } = unify ( schematic , %{ name : &quot;George Hill&quot; , team : &quot;Indiana Pacers&quot; } ) iex&gt; { :error , &quot;expected either a map or a map&quot; } = unify ( schematic , %{ name : &quot;NBA&quot; , sport : &quot;basketball&quot; } ) With a function When called with a function, the input is passed as the only parameter. This can be used to dispach to a specific schematic. This is a performance optimization, as you can dispatch to a specific schematic rather than traversing all of them. iex&gt; schematic = oneof ( fn ...&gt; %{ type : &quot;team&quot; } -&gt; map ( %{ name : str ( ) , league : str ( ) } ) ...&gt; %{ type : &quot;player&quot; } -&gt; map ( %{ name : str ( ) , team : str ( ) } ) ...&gt; _ -&gt; { :error , &quot;expected either a player or a team&quot; } ...&gt; end ) iex&gt; { :ok , %{ name : &quot;Indiana Pacers&quot; , league : &quot;NBA&quot; } } = unify ( schematic , %{ type : &quot;team&quot; , name : &quot;Indiana Pacers&quot; , league : &quot;NBA&quot; } ) iex&gt; { :ok , %{ name : &quot;George Hill&quot; , team : &quot;Indiana Pacers&quot; } } = unify ( schematic , %{ type : &quot;player&quot; , name : &quot;George Hill&quot; , team : &quot;Indiana Pacers&quot; } ) iex&gt; { :error , &quot;expected either a player or a team&quot; } = unify ( schematic , %{ name : &quot;NBA&quot; , sport : &quot;basketball&quot; } )","ref":"Schematic.html#oneof/1"},{"type":"function","title":"Schematic.optional/1","doc":"See map/1 for examples and explanation.","ref":"Schematic.html#optional/1"},{"type":"function","title":"Schematic.optional/2","doc":"Specifies an optional key and also a default value in the case the key is not present. See map/1 for more examples and explanation.","ref":"Schematic.html#optional/2"},{"type":"function","title":"Schematic.raw/2","doc":"A utility for creating custom schematics. The raw/1 schematic is useful for creating schematics that unify the values of the inputs, rather than just the shape. Options :message - a custom error message. Defaults to &quot;is invalid&quot; . :transformer - a function that takes the input and the unification direction and must return the desired value. Defaults to fn input, _dir -&gt; input end . Basic Usage iex&gt; schematic = all ( [ int ( ) , raw ( fn i -&gt; i &gt; 10 end , message : &quot;must be greater than 10&quot; ) ] ) iex&gt; { :ok , 11 } = unify ( schematic , 11 ) iex&gt; { :error , [ &quot;must be greater than 10&quot; ] } = unify ( schematic , 9 ) Advanced Usage If your data requires different validations for unification and dumping, then you can pass a 2-arity function (instead of a 1-arity function) and the second parameter will be the direction. This concept also applies to the :transform option. iex&gt; schematic = ...&gt; raw ( ...&gt; fn ...&gt; n , :to -&gt; is_list ( n ) and length ( n ) == 3 ...&gt; n , :from -&gt; is_tuple ( n ) and tuple_size ( n ) == 3 ...&gt; end , ...&gt; message : &quot;must be a tuple of size 3&quot; , ...&gt; transform : fn ...&gt; input , :to -&gt; ...&gt; List . to_tuple ( input ) ...&gt; input , :from -&gt; ...&gt; Tuple . to_list ( input ) ...&gt; end ...&gt; ) iex&gt; { :ok , { &quot;one&quot; , &quot;two&quot; , 3 } } = unify ( schematic , [ &quot;one&quot; , &quot;two&quot; , 3 ] ) iex&gt; { :error , &quot;must be a tuple of size 3&quot; } = unify ( schematic , [ &quot;not&quot; , &quot;big&quot; ] ) iex&gt; { :ok , [ &quot;one&quot; , &quot;two&quot; , 3 ] } = dump ( schematic , { &quot;one&quot; , &quot;two&quot; , 3 } )","ref":"Schematic.html#raw/2"},{"type":"function","title":"Schematic.schema/3","doc":"Specifies a map/1 schematic that is then hydrated into a struct. Works the same as the map/1 schematic, but can also automatically transform all keys from string keys to atom keys if a key conversion is not already specified. This can be disabled by passing the convert: false option Since this schematic hydrates a struct, it is also only capable of having atom keys in the output, whereas a normal map can have arbitrary terms as the key. iex&gt; schematic = ...&gt; schema ( HTTPRequest , %{ ...&gt; method : oneof ( [ &quot;POST&quot; , &quot;PUT&quot; , &quot;PATCH&quot; ] ) , ...&gt; body : str ( ) ...&gt; } ) iex&gt; { :ok , % HTTPRequest { method : &quot;POST&quot; , body : ~s|{&quot;name&quot;: &quot;Peter&quot;}| } } = unify ( schematic , %{ &quot;method&quot; =&gt; &quot;POST&quot; , &quot;body&quot; =&gt; ~s|{&quot;name&quot;: &quot;Peter&quot;}| } ) iex&gt; { :ok , %{ &quot;method&quot; =&gt; &quot;POST&quot; , &quot;body&quot; =&gt; ~s|{&quot;name&quot;: &quot;Peter&quot;}| } } = dump ( schematic , % HTTPRequest { method : &quot;POST&quot; , body : ~s|{&quot;name&quot;: &quot;Peter&quot;}| } ) Options :convert - Automatically convert string keys to atom keys. Defaults to true .","ref":"Schematic.html#schema/3"},{"type":"function","title":"Schematic.str/0","doc":"Specifies that the data is a string or a specific string. Usage Any string. iex&gt; schematic = str ( ) iex&gt; { :ok , &quot;hi!&quot; } = unify ( schematic , &quot;hi!&quot; ) iex&gt; { :error , &quot;expected a string&quot; } = unify ( schematic , :boom ) A string literal. iex&gt; schematic = &quot;I 💜 Elixir&quot; iex&gt; { :ok , &quot;I 💜 Elixir&quot; } = unify ( schematic , &quot;I 💜 Elixir&quot; ) iex&gt; { :error , ~s|expected &quot;I 💜 Elixir&quot;| } = unify ( schematic , &quot;I love Ruby&quot; )","ref":"Schematic.html#str/0"},{"type":"function","title":"Schematic.tuple/2","doc":"Specifies that the data is a tuple of the given length where each element unifies to the schematic in the same position. Usage iex&gt; schematic = tuple ( [ str ( ) , int ( ) ] ) iex&gt; { :ok , { &quot;one&quot; , 2 } } = unify ( schematic , { &quot;one&quot; , 2 } ) iex&gt; { :error , &quot;expected a tuple of {a string, an integer}&quot; } = unify ( schematic , { 1 , &quot;two&quot; } ) Options :from - Either :tuple or :list . Defaults to :tuple . iex&gt; schematic = tuple ( [ str ( ) , int ( ) ] , from : :list ) iex&gt; { :ok , { &quot;one&quot; , 2 } } = unify ( schematic , [ &quot;one&quot; , 2 ] ) iex&gt; { :error , &quot;expected a list of {a string, an integer}&quot; } = unify ( schematic , [ 1 , &quot;two&quot; ] )","ref":"Schematic.html#tuple/2"},{"type":"function","title":"Schematic.unify/2","doc":"Unify external data with your internal data structures. See all the other functions for information on how to create schematics.","ref":"Schematic.html#unify/2"},{"type":"type","title":"Schematic.lazy_schematic/0","doc":"A lazy reference to a schematic, used to define recursive schematics.","ref":"Schematic.html#t:lazy_schematic/0"},{"type":"type","title":"Schematic.literal/0","doc":"A literal schematic.","ref":"Schematic.html#t:literal/0"},{"type":"type","title":"Schematic.map_blueprint/0","doc":"The blueprint used to specify a map schematic.","ref":"Schematic.html#t:map_blueprint/0"},{"type":"type","title":"Schematic.map_blueprint_key/0","doc":"Map blueprint key.","ref":"Schematic.html#t:map_blueprint_key/0"},{"type":"type","title":"Schematic.map_blueprint_value/0","doc":"Map blueprint value.","ref":"Schematic.html#t:map_blueprint_value/0"},{"type":"type","title":"Schematic.schema_blueprint/0","doc":"The blueprint used to specify a schema schematic.","ref":"Schematic.html#t:schema_blueprint/0"},{"type":"type","title":"Schematic.schema_blueprint_key/0","doc":"Schema blueprint key.","ref":"Schematic.html#t:schema_blueprint_key/0"},{"type":"type","title":"Schematic.schema_blueprint_value/0","doc":"Schema blueprint value.","ref":"Schematic.html#t:schema_blueprint_value/0"},{"type":"type","title":"Schematic.t/0","doc":"The Schematic data structure. This data structure is meant to be opaque to the user, but you can create your own for super niche use cases. But backwards compatiblility of this data structure is not guaranteed.","ref":"Schematic.html#t:t/0"},{"type":"module","title":"Schematic.OptionalKey","doc":"","ref":"Schematic.OptionalKey.html"},{"type":"opaque","title":"Schematic.OptionalKey.t/0","doc":"","ref":"Schematic.OptionalKey.html#t:t/0"}]